/***************************************************************************************************
Konverter zur Umwandlung von M-Link Telemetriedaten in das FrSky Protokoll
****************************************************************************************************
Controller:     ATmega328P
Taktfrequenz:   16 MHz
Entwickler:     Reinhardt Werbik
Version:        1.00
Datum:          30.10.2017
****************************************************************************************************
Der Konverter empfngt Telemetriedaten von einem M-Link Sendemodul und konvertiert diese in das von
FrSky benutzte Telemetrieformat (D-Protokoll fr Sensor-Hub).
(Der Konverter ist in einem separaten Dokument detailliert beschrieben.)

Im Hauptmodul ist der bergeordnete zeitliche Ablauf implementiert, wobei das Senden der FrSky Daten
mit dem Empfang der M-Link Telemetriedaten synchronisiert wird, um Datenkollisionen zu vermeiden.

Das M-Link Modul signalisiert nach dem Einschalten immer kurzzeitig gltige Telemetriedaten.
(auch bei ausgeschaltetem Empfnger)
Daher wird nach dem Start eine definierte Anzahl von M-Link Datenpaketen ignoriert,
damit nicht vorzeitig die Telemetrie-bertragung aktiviert wird.

Nach einer definierten Anzahl aufeinander folgender M-Link Datenpakete mit gltigen Telemetriedaten
wird die Telemetrie-bertragung aktiviert.
Nach einer definierten Anzahl aufeinander folgender M-Link Datenpakete ohne gltige Telemetriedaten
wird die Telemetrie-bertragung deaktiviert, wobei alle Sensordaten neu initialisiert werden.
Dadurch ist sichergestellt, dass nach Ausschalten des Empfngers und erneutem Einschalten mit
teilweise entfernten Sensoren die entsprechenden Parameter nicht mehr bertragen werden.

Der Sendetakt wird erst nach der vollstndigen Verarbeitung eines M-Link Datenpakets gestartet.
Er wird vor dem nchsten M-Link Datenpaket (sptestens nach vier gesendeten FrSky Datenpaketen)
wieder angehalten, so dass dieses empfangen werden kann, ohne durch Interrupts des SW UARTs zum
Senden der FrSky Daten beeintrchtigt zu werden.

Die Anzahl der pro M-Link Datenpaket maximal zu sendenden FrSky Datenpakete reduziert sich auf drei,
wenn Fast Response aktiviert ist, da dann der zeitliche Abstand zwischen zwei M-Link Datenpaketen
kleiner ist (unterschiedliche Wiederholrate der M-Link Daten).

Eine Status-LED zeigt mit verschiedenen Blinkfrequenzen an,
ob Daten vom M-Link Modul empfangen werden, und ob diese gltige Telemetrie-Werte beinhalten.
Der Start bzw. Reset des Konverters wird durch eine spezielle Blinksequenz der Status-LED angezeigt.
Das Signal fr die Status-LED wird an zwei verschiedenen Pins ausgegeben, um sowohl die LED auf dem
Arduino Board, als auch eine zustzliche externe LED ansteuern zu knnen.

Timer 1 dient nach der Initialisierung zur Erzeugung einer Blinksequenz fr die Status-LED.
Der Timer wird dazu im Normal Mode mit Prescaler 1024 betrieben.

Danach erzeugt Timer 1 mit dem Compare Match A Interrupt den Sendetakt fr die FrSky Daten.
Der Timer wird dazu im CTC Mode mit Prescaler 256 (Vergleichswert A = 1249) betrieben.
(Taktzeit = (1249 + 1) * 256 / 16 MHz = 20 ms)

Der Watchdog Timer wird verwendet, um den Konverter neu zu starten, wenn fr lngere Zeit
keine M-Link Datenpakete mehr ankommen.
Dadurch ist sichergestellt, dass bei einem Neustart des M-Link Moduls ohne Ausschalten des
Senders (z.B. nach Wechsel des Modellspeichers) auch der Konverter neu gestartet wird.
****************************************************************************************************
nderungen:

V 1.00:
- Erste konfigurierte Version
***************************************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <stdbool.h>
#include <avr/wdt.h>

#include "../include/mlink_data.h"
#include "../include/mlink_interface.h"
#include "../include/frsky_data.h"
#include "../include/frsky_interface.h"

// Konfigurations-Parameter

#define START_COUNT     32      // Anzahl nach dem Start ignorierter M-Link Datenpakete
#define THRESHOLD       8       // Schwelle fr Telemetrie aktiv/passiv (Anzahl der Datenpakete)

// Status-LED

#define LED_PORT    PORTB       // Port B verwenden
#define LED_DDR     DDRB        // Port B DDR
#define LED1        PINB0       // Pin B0 fr externe Status-LED
#define LED2        PINB5       // Pin B5 fr Status-LED auf Arduino Board

#define LED_ON      LED_PORT |= (1 << LED1) | (1 << LED2)       // Status-LED einschalten
#define LED_OFF     LED_PORT &= ~((1 << LED1) | (1 << LED2))    // Status-LED ausschalten

//========== Variablen

static volatile uint8_t frsky_count = 0;    // Zhler fr gesendete FrSky Datenpakete
static volatile uint8_t tx_packages = 4;    // Max. Anzahl von FrSky Datenpaketen pro M-Link Paket

//========== Interrupt Service Routinen

ISR(TIMER1_COMPA_vect)          // Sendetakt fr FrSky Telemetriedaten
{
	handle_frsky_tx_request(frsky_count);
	frsky_count ++;
	if (frsky_count == tx_packages)
	{
		TCCR1B = 0;     // Sendetakt anhalten
	}
}

//========== Hauptprogramm

int main(void)
{
	//===== Variablen

	uint8_t mlink_count = 0;            // Zhler fr empfangene M-Link Datenpakete
	uint8_t led_count = 0;              // Zhler zur Steuerung der Status-LED
	bool telemetry_active = false;      // Flag fr aktive Telemetrie

	//===== Initialisierung

	DDRB = 0x00;        // Port B als Input konfigurieren
	PORTB = 0xFF;       // Pull-Ups aktivieren
	DDRC = 0x00;        // Port C als Input konfigurieren
	PORTC = 0xFF;       // Pull-Ups aktivieren
	DDRD = 0x00;        // Port D als Input konfigurieren
	PORTD = 0xFF;       // Pull-Ups aktivieren

	// Status-LED
	LED_DDR |= (1 << LED1) | (1 << LED2);   // LED Pins als Ausgang konfigurieren
	LED_OFF;

	init_mlink_interface();
	init_frsky_interface();
	init_msb_data();
	init_frsky_data();

	sei();      // Interrupts freigeben

	//===== Programmbeginn

	// Zunchst wird eine spezielle Blinksequenz fr die Status-LED erzeugt (dreimal lang/kurz/kurz).
	// Timer 1 luft dazu dreimal fr ca. 1,2 s, das LED-Signal wird aus dem Timer-Wert erzeugt.

	uint8_t i;

	for (i = 0; i < 3; i++)
	{
		TCNT1 = 0;                              // Timer zurcksetzen
		TCCR1B = (1 << CS12) | (1 << CS10);     // Takt starten (Prescaler 1024)
		while (TCNT1 < 18750)                   // und ca. 1,2 s laufen lassen
		{
			led_count = (TCNT1 >> 10);      // obere 6 Bits des Timer-Werts extrahieren
			switch (led_count)
			{
				case 4:
					LED_ON;
					break;
				case 8:
					LED_OFF;
					break;
				case 12:
					LED_ON;
					break;
				case 13:
					LED_OFF;
					break;
				case 15:
					LED_ON;
					break;
				case 16:
					LED_OFF;
			}
		}
	}
	TCCR1B = 0;         // Takt anhalten
	TCNT1 = 0;          // Timer zurcksetzen
	led_count = 0;      // Hilfszhler fr Status-LED zurcksetzen

	// Timer 1 fr Sendetakt konfigurieren
	OCR1A = 1249;               // Vergleichswert A setzen
	TIMSK1 = (1 << OCIE1A);     // Compare Match A Interrupt freigeben

	// Nach dem Start wird eine definierte Anzahl von M-Link Datenpaketen ignoriert.
	// Danach wird die max. Anzahl von FrSky Datenpaketen nach einem M-Link Datenpaket reduziert,
	// falls Fast Response aktiviert ist.

	while (mlink_count < START_COUNT)
	{
		if (new_mlink_data())
		{
			LED_ON;
			mlink_count ++;
		}
	}
	mlink_count = 0;

	if (fast_response_enabled())
	{
		tx_packages = 3;    // nur max. drei FrSky Datenpakete bei Fast Response
	}

	wdt_enable(WDTO_4S);    // Watchdog Timer mit 4 s Timeout aktivieren

	// Jetzt beginnt die normale Verarbeitung von M-Link Datenpaketen.

	while(1)
	{
		if (new_mlink_data())
		{
			if (telemetry_active)
			{
				if (mlink_data_valid())
				{
					process_mlink_data();
					update_frsky_data();
					mlink_count = 0;
				}
				else
				{
					mlink_count ++;
					if (mlink_count == THRESHOLD)
					{
						telemetry_active = false;
						init_msb_data();
						init_frsky_data();
					}
				}

				// Der FrSky Sendetakt wird gestartet, wobei der Timer so vorgeladen wird,
				// dass der erste Interrupt nach ca. 4 ms auftritt.

				frsky_count = 0;                        // Taktzhler zurcksetzen
				TCNT1 = 1000;                           // Timer vorladen
				TCCR1B = (1 << WGM12) | (1 << CS12);    // Sendetakt starten

				// Status-LED ansteuern (langsames Blinken)
				if (led_count & 0x08)
				{
					LED_ON;
				}
				else
				{
					LED_OFF;
				}
			}
			else
			{
				if (mlink_data_valid())
				{
					mlink_count ++;
					if (mlink_count == THRESHOLD)
					{
						telemetry_active = true;
					}
				}
				else
				{
					mlink_count = 0;
				}
				// Status-LED ansteuern (schnelles Blinken)
				if (led_count & 0x01)
				{
					LED_ON;
				}
				else
				{
					LED_OFF;
				}
			}

			led_count ++;
			wdt_reset();
		}
	}
}