/***************************************************************************************************
Daten-Modul fr FrSky Sensordaten
****************************************************************************************************
Controller:     -
Taktfrequenz:   -
Entwickler:     Reinhardt Werbik
Version:        1.00
Datum:          30.10.2017
****************************************************************************************************
Dieses Modul implementiert die Konvertierung der MSB Daten in die entsprechenden FrSky Daten.
Wie die MSB Sensorwerte umgewandelt werden, ist in einem separaten Dokument ausfhrlich beschrieben.

Die Spannungen A1 und A2, sowie der RSSI Wert werden in einzelnen Variablen gespeichert.
Diese Parameter werden im RVLQ (Remote Voltage and Link Quality) Frame bertragen.

Die User Daten werden in einem entsprechenden Daten-Array gespeichert.
Der Array-Index entspricht dabei der MSB Adresse des zugehrigen Sensorwertes.
Jeder Datensatz beinhaltet eine eindeutige ID, den Wert des Parameters und ein Gltigkeits-Flag.
Die ID ergibt sich aus der MSB Adresse des zugehrigen Sensorwertes, unabhngig von der Werteklasse.
Ein zustzliches Flag zeigt an, ob der Datensatz aktualisiert wurde.

Die Alarm-Flags aller gltigen MSB Parameter werden in einer separaten Variablen gespeichert.

Das Modul verarbeitet Anforderungen zum Senden eines FrSky Daten-Frames.
Ein bestimmter User Data Frame wird gesendet, wenn der zugehrige Sensorwert aktualisiert wurde.
Der RVLQ Frame wird immer gesendet, unabhngig davon, ob die darin enthaltenen Daten gltig sind.
Die Alarm Flags werden mit der niedrigsten Prioritt bertragen.

Der Code in diesem Modul ist unabhngig vom verwendeten Controller und der Taktfrequenz.
****************************************************************************************************
nderungen:

V 1.00:
- Erste konfigurierte Version
***************************************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <stdbool.h>

#include "mlink_data.h"
#include "frsky_data.h"
#include "frsky_interface.h"

// IDs fr User Daten (Low Nibble = MSB Adresse)

#define ID_MSB_00   0x20
#define ID_MSB_01   0x31
#define ID_MSB_02   0x32
#define ID_MSB_03   0x33
#define ID_MSB_04   0x34
#define ID_MSB_05   0x35
#define ID_MSB_06   0x36
#define ID_MSB_07   0x37
#define ID_MSB_08   0x38
#define ID_MSB_09   0x29
#define ID_MSB_10   0x2A
#define ID_MSB_11   0x2B
#define ID_MSB_12   0x2C
#define ID_MSB_13   0x2D
#define ID_MSB_14   0x2E
#define ID_MSB_15   0x2F

#define ID_ALARMS   0x0F    // MSB Alarm-Flags

// Vordefinierte MSB Adressen der RVLQ Parameter

#define ADDR_A1     0       // A1 Spannung
#define ADDR_A2     2       // A2 Spannung
#define ADDR_RSSI   1       // RSSI Wert (LQI auf dem MSB)

//========== Datentypen

typedef struct
{
	uint8_t id;         // Daten ID
	int16_t value;      // Sensorwert
	bool valid;         // Gltigkeit des Sensorwertes
	bool updated;       // zeigt an, ob der Datensatz aktualisiert wurde
} user_data_t;

//========== Variablen

static uint8_t a1_voltage = 0;          // A1 Spannung (normalerweise die Empfngerspannung)
static uint8_t a2_voltage = 0;          // A2 Spannung (eine weitere Spannung bis 25,5 V)
static uint8_t rssi = 0;                // RSSI Wert (LQI auf dem MSB)

static user_data_t user_data[16];       // User Daten (Index = MSB Adresse)

static uint16_t alarm_flags = 0;        // MSB Alarm-Flags (Bit-Position = MSB-Adresse)

static bool msb_used[16];       // zeigt fr jede MSB Adresse an, ob sie bereits verwendet wird

//========== Interne Prototypen

uint8_t clipped_value(int16_t value);

//========== Interne Funktionen

uint8_t clipped_value(int16_t value)
{
	if (value > 255)
	{
		return 255;     // auf 8 Bit Wert begrenzen
	}
	else
	{
		if (value < 0)
		{
			return 0;
		}
		else
		{
			return value;
		}
	}
}

//========== Interface Funktionen

void init_frsky_data(void)
{
	uint8_t i;

	for (i = 0; i < 16; i++)
	{
		user_data[i].value = 0;
		user_data[i].valid = false;
		user_data[i].updated = false;
	}

	user_data[0].id  = ID_MSB_00;
	user_data[1].id  = ID_MSB_01;
	user_data[2].id  = ID_MSB_02;
	user_data[3].id  = ID_MSB_03;
	user_data[4].id  = ID_MSB_04;
	user_data[5].id  = ID_MSB_05;
	user_data[6].id  = ID_MSB_06;
	user_data[7].id  = ID_MSB_07;
	user_data[8].id  = ID_MSB_08;
	user_data[9].id  = ID_MSB_09;
	user_data[10].id = ID_MSB_10;
	user_data[11].id = ID_MSB_11;
	user_data[12].id = ID_MSB_12;
	user_data[13].id = ID_MSB_13;
	user_data[14].id = ID_MSB_14;
	user_data[15].id = ID_MSB_15;
}

void update_frsky_data(void)
{
	uint8_t i;

	for (i = 0; i < 16; i++)
	{
		msb_used[i] = false;
	}

	// Die Daten fr den RVLQ Frame werden berechnet, wenn auf den zugeordneten MSB Adressen
	// gltige Sensorwerte mit den passenden Werteklassen bertragen werden.
	// (Diese Adressen knnen auch fr beliebige andere Parameter verwendet werden.)

	if ((msb_data[ADDR_A1].class == MSB_VOLTAGE) && msb_data[ADDR_A1].valid)    // A1 Spannung
	{
		a1_voltage = clipped_value(msb_data[ADDR_A1].value);    // auf 25,5 V begrenzen (8 Bit)
		msb_used[ADDR_A1] = true;
		if (msb_data[ADDR_A1].updated)
		{
			msb_data[ADDR_A1].updated = false;
		}
	}
	if ((msb_data[ADDR_A2].class == MSB_VOLTAGE) && msb_data[ADDR_A2].valid)    // A2 Spannung
	{
		a2_voltage = clipped_value(msb_data[ADDR_A2].value);    // auf 25,5 V begrenzen (8 Bit)
		msb_used[ADDR_A2] = true;
		if (msb_data[ADDR_A2].updated)
		{
			msb_data[ADDR_A2].updated = false;
		}
	}
	if ((msb_data[ADDR_RSSI].class == MSB_LQI) && msb_data[ADDR_RSSI].valid)    // RSSI
	{
		rssi = msb_data[ADDR_RSSI].value;
		msb_used[ADDR_RSSI] = true;
		if (msb_data[ADDR_RSSI].updated)
		{
			msb_data[ADDR_RSSI].updated = false;
		}
	}

	// Die User Daten (mit IDs gem MSB Adresse) werden berechnet.

	for (i = 0; i < 16; i++)
	{
		if (!msb_used[i] && msb_data[i].valid)
		{
			if (msb_data[i].class == MSB_RPM)
			{
				// Fr Drehzahlen gibt es zwei mgliche Auflsungen (10 bzw. 100 1/min).
				// Beide Formate werden auf 1/min mit positivem Vorzeichen umgerechnet.

				uint16_t rpm = 0;

				if (msb_data[i].value < 0)
				{
					// MSB Auflsung 10 1/min (durch negatives Vorzeichen gekennzeichnet)
					rpm = -msb_data[i].value * 10;
				}
				else
				{
					// MSB Auflsung 100 1/min (durch positives Vorzeichen gekennzeichnet)
					rpm = msb_data[i].value * 100;
				}
				if (rpm > 32767)    // Begrenzung auf maximalen positiven Wert fr User Daten
				{
					rpm = 32767;
				}
				user_data[i].value = rpm;
			}
			else
			{
				// Alle anderen Sensorwerte werden einfach unverndert durchgereicht.
				user_data[i].value = msb_data[i].value;
			}
			user_data[i].valid = true;
			msb_used[i] = true;         // verwendete MSB Adresse kennzeichnen
			if (msb_data[i].updated)    // ggf. MSB Update-Flag bernehmen und dann zurcksetzen
			{
				user_data[i].updated = true;
				msb_data[i].updated = false;
			}
		}
		else
		{
			user_data[i].valid = false;
		}
	}

	// Die MSB Alarm-Flags werden abgefragt.

	alarm_flags = 0;
	for (i = 0; i < 16; i++)
	{
		if (msb_data[i].valid && msb_data[i].alarm)
		{
			alarm_flags |= (1 << i);
		}
	}
}

void handle_frsky_tx_request(uint8_t count)
{
	static bool rvlq_sent = false;      // zeigt an ob bereits ein RVLQ Frame gesendet wurde.
	static bool alarms_sent = false;    // zeigt an ob bereits die Alarm Flags gesendet wurden

	uint8_t i;

	// Die Flags fr RVLQ Frame und Alarm Flags werden vor dem ersten Frame zurckgesetzt.
	if (count == 0)
	{
		rvlq_sent = false;
		alarms_sent = false;
	}

	// Frame fr User Daten (falls ein Parameter aktualisiert wurde)
	for (i = 0; i < 16; i++)
	{
		if (user_data[i].updated)
		{
			send_frsky_user_frame(user_data[i].id, user_data[i].value);
			user_data[i].updated = false;
			return;
		}
	}
	if (!(rvlq_sent))
	{
		// RVLQ Frame (wird in jeder Sequenz maximal einmal gesendet)
		send_frsky_rvlq_frame(a1_voltage, a2_voltage, rssi);
		rvlq_sent = true;
		return;
	}
	if (!(alarms_sent))
	{
		// Frame fr MSB Alarm-Flags (wird in jeder Sequenz maximal einmal gesendet)
		send_frsky_user_frame(ID_ALARMS, alarm_flags);
		alarms_sent = true;
	}
}