/***************************************************************************************************
Daten-Modul fr FrSky Sensordaten (Variante mit Benutzung vordefinierter FrSky IDs)
****************************************************************************************************
Controller:     -
Taktfrequenz:   -
Entwickler:     Reinhardt Werbik
Version:        1.00
Datum:          30.10.2017
****************************************************************************************************
Dieses Modul implementiert die Konvertierung der MSB Daten in die entsprechenden FrSky Daten.
Wie die MSB Sensorwerte umgewandelt werden, ist in einem separaten Dokument ausfhrlich beschrieben.

Die Spannungen A1 und A2, sowie der RSSI Wert werden in einzelnen Variablen gespeichert.
Diese Parameter werden im RVLQ (Remote Voltage and Link Quality) Frame bertragen.

Die FrSky User Daten mit vordefinierten IDs werden in einem entsprechenden Daten-Array gespeichert.
Jeder Datensatz beinhaltet eine eindeutige ID, den Wert des Parameters und ein Gltigkeits-Flag.
Ein zustzliches Flag zeigt an, ob der Datensatz aktualisiert wurde.

Ein weiteres Daten-Array dient zum Speichern der zustzlichen User Daten mit IDs gem MSB Adresse.
Die ID ergibt sich aus der MSB Adresse des zugehrigen Sensorwertes, unabhngig von der Werteklasse.
Der Array-Index entspricht hier der MSB Adresse des zugehrigen Sensorwertes.

Die Alarm-Flags aller gltigen MSB Parameter werden in einer separaten Variablen gespeichert.

Das Modul verarbeitet Anforderungen zum Senden eines FrSky Daten-Frames.
Ein bestimmter User Data Frame wird gesendet, wenn der zugehrige Sensorwert aktualisiert wurde.
Der RVLQ Frame wird immer gesendet, unabhngig davon, ob die darin enthaltenen Daten gltig sind.
Die Alarm Flags werden mit der niedrigsten Prioritt bertragen.

Der Code in diesem Modul ist unabhngig vom verwendeten Controller und der Taktfrequenz.
****************************************************************************************************
nderungen:

V 1.00:
- Erste konfigurierte Version
***************************************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <stdbool.h>

#include "mlink_data.h"
#include "frsky_data.h"
#include "frsky_interface.h"

// IDs fr FrSky User Daten

#define ID_VERT_SPEED       0x30    // Steigrate
#define ID_VOLTAGE          0x39    // Spannung
#define ID_CURRENT          0x28    // Strom
#define ID_TEMP1            0x02    // Temperatur 1
#define ID_TEMP2            0x05    // Temperatur 2

#define ID_MSB_00           0x20    // zustzliche User Daten (Low Nibble = MSB Adresse)
#define ID_MSB_01           0x31
#define ID_MSB_02           0x32
#define ID_MSB_03           0x33
#define ID_MSB_04           0x34
#define ID_MSB_05           0x35
#define ID_MSB_06           0x36
#define ID_MSB_07           0x37
#define ID_MSB_08           0x38
#define ID_MSB_09           0x29
#define ID_MSB_10           0x2A
#define ID_MSB_11           0x2B
#define ID_MSB_12           0x2C
#define ID_MSB_13           0x2D
#define ID_MSB_14           0x2E
#define ID_MSB_15           0x2F

#define ID_ALARMS           0x0F    // MSB Alarm-Flags

// Indizes fr User Daten mit FrSky ID (mssen lckenlos und eindeutig sein)

#define INDEX_VERT_SPEED    0
#define INDEX_VOLTAGE       1
#define INDEX_CURRENT       2
#define INDEX_TEMP1         3
#define INDEX_TEMP2         4

#define USER_COUNT  5       // Anzahl verschiedener User Data Frames mit FrSky ID

// Vordefinierte MSB Adressen der RVLQ Parameter

#define ADDR_A1     0       // A1 Spannung
#define ADDR_A2     2       // A2 Spannung
#define ADDR_RSSI   1       // RSSI Wert (LQI auf dem MSB)

//========== Datentypen

typedef struct
{
	uint8_t id;         // Daten ID
	int16_t value;      // Sensorwert
	bool valid;         // Gltigkeit des Sensorwertes
	bool updated;       // zeigt an, ob der Datensatz aktualisiert wurde
} user_data_t;

//========== Variablen

static uint8_t a1_voltage = 0;              // A1 Spannung (normalerweise die Empfngerspannung)
static uint8_t a2_voltage = 0;              // A2 Spannung (eine weitere Spannung bis 25,5 V)
static uint8_t rssi = 0;                    // RSSI Wert (LQI auf dem MSB)

static user_data_t user_data[USER_COUNT];   // FrSky User Daten (Index = Prioritts-Reihenfolge)
static user_data_t extra_data[16];          // zustzliche User Daten (Index = MSB Adresse)

static uint16_t alarm_flags = 0;            // MSB Alarm-Flags (Bit-Position = MSB-Adresse)

static bool msb_used[16];       // zeigt fr jede MSB Adresse an, ob sie bereits verwendet wird

//========== Interne Prototypen

uint16_t unsigned_value(int16_t value);
uint8_t clipped_value(int16_t value);
void set_user_data_valid(uint8_t index, uint8_t addr);

//========== Interne Funktionen

uint16_t unsigned_value(int16_t value)
{
	if (value < 0)
	{
		return 0;
	}
	else
	{
		return value;
	}
}

uint8_t clipped_value(int16_t value)
{
	if (value > 255)
	{
		return 255;     // auf 8 Bit Wert begrenzen
	}
	else
	{
		return unsigned_value(value);
	}
}

void set_user_data_valid(uint8_t index, uint8_t addr)
{
	user_data[index].valid = true;
	msb_used[addr] = true;              // verwendete MSB Adresse kennzeichnen
	if (msb_data[addr].updated)         // ggf. MSB Update-Flag bernehmen und dann zurcksetzen
	{
		user_data[index].updated = true;
		msb_data[addr].updated = false;
	}
}

//========== Interface Funktionen

void init_frsky_data(void)
{
	uint8_t i;

	for (i = 0; i < USER_COUNT; i++)
	{
		user_data[i].value = 0;
		user_data[i].valid = false;
		user_data[i].updated = false;
	}

	// IDs fr FrSky User Daten zuordnen
	user_data[INDEX_VERT_SPEED].id    = ID_VERT_SPEED;
	user_data[INDEX_VOLTAGE].id       = ID_VOLTAGE;
	user_data[INDEX_CURRENT].id       = ID_CURRENT;
	user_data[INDEX_TEMP1].id         = ID_TEMP1;
	user_data[INDEX_TEMP2].id         = ID_TEMP2;

	for (i = 0; i < 16; i++)
	{
		extra_data[i].value = 0;
		extra_data[i].valid = false;
		extra_data[i].updated = false;
	}

	// IDs fr zustzliche User Daten zuordnen
	extra_data[0].id  = ID_MSB_00;
	extra_data[1].id  = ID_MSB_01;
	extra_data[2].id  = ID_MSB_02;
	extra_data[3].id  = ID_MSB_03;
	extra_data[4].id  = ID_MSB_04;
	extra_data[5].id  = ID_MSB_05;
	extra_data[6].id  = ID_MSB_06;
	extra_data[7].id  = ID_MSB_07;
	extra_data[8].id  = ID_MSB_08;
	extra_data[9].id  = ID_MSB_09;
	extra_data[10].id = ID_MSB_10;
	extra_data[11].id = ID_MSB_11;
	extra_data[12].id = ID_MSB_12;
	extra_data[13].id = ID_MSB_13;
	extra_data[14].id = ID_MSB_14;
	extra_data[15].id = ID_MSB_15;
}

void update_frsky_data(void)
{
	uint8_t i;

	for (i = 0; i < USER_COUNT; i++)
	{
		user_data[i].valid = false;
	}
	for (i = 0; i < 16; i++)
	{
		msb_used[i] = false;
	}

	// Die Daten fr den RVLQ Frame werden berechnet, wenn auf den zugeordneten MSB Adressen
	// gltige Sensorwerte mit den passenden Werteklassen bertragen werden.
	// (Diese Adressen knnen auch fr beliebige andere Parameter verwendet werden.)

	if ((msb_data[ADDR_A1].class == MSB_VOLTAGE) && msb_data[ADDR_A1].valid)    // A1 Spannung
	{
		a1_voltage = clipped_value(msb_data[ADDR_A1].value);    // auf 25,5 V begrenzen (8 Bit)
		msb_used[ADDR_A1] = true;
		if (msb_data[ADDR_A1].updated)
		{
			msb_data[ADDR_A1].updated = false;
		}
	}
	if ((msb_data[ADDR_A2].class == MSB_VOLTAGE) && msb_data[ADDR_A2].valid)    // A2 Spannung
	{
		a2_voltage = clipped_value(msb_data[ADDR_A2].value);    // auf 25,5 V begrenzen (8 Bit)
		msb_used[ADDR_A2] = true;
		if (msb_data[ADDR_A2].updated)
		{
			msb_data[ADDR_A2].updated = false;
		}
	}
	if ((msb_data[ADDR_RSSI].class == MSB_LQI) && msb_data[ADDR_RSSI].valid)    // RSSI
	{
		rssi = msb_data[ADDR_RSSI].value;
		msb_used[ADDR_RSSI] = true;
		if (msb_data[ADDR_RSSI].updated)
		{
			msb_data[ADDR_RSSI].updated = false;
		}
	}

	// Die FrSky User Daten mit vordefinierten IDs werden berechnet.

	for (i = 0; i < 16; i++)
	{
		if (!msb_used[i] && msb_data[i].valid)
		{
			switch (msb_data[i].class)
			{
				case MSB_VOLTAGE:                               // ===== Spannung
					if (!user_data[INDEX_VOLTAGE].valid)
					{
						user_data[INDEX_VOLTAGE].value = unsigned_value(msb_data[i].value);
						set_user_data_valid(INDEX_VOLTAGE, i);
					}
					break;
				case MSB_CURRENT:                               // ===== Strom
					if (!user_data[INDEX_CURRENT].valid)
					{
						user_data[INDEX_CURRENT].value = unsigned_value(msb_data[i].value);
						set_user_data_valid(INDEX_CURRENT, i);
					}
					break;
				case MSB_TEMP:
					// Die MSB Auflsung von 0,1 C wird beibehalten.
					// Der Sensorwert muss daher im Sender entsprechend konfiguriert werden.
					if (!user_data[INDEX_TEMP1].valid)          // ===== Temperatur 1
					{
						user_data[INDEX_TEMP1].value = msb_data[i].value;
						set_user_data_valid(INDEX_TEMP1, i);
					}
					else
					{
						if (!user_data[INDEX_TEMP2].valid)      // ===== Temperatur 2
						{
							user_data[INDEX_TEMP2].value = msb_data[i].value;
							set_user_data_valid(INDEX_TEMP2, i);
						}
					}
					break;
				case MSB_VERT_SPEED:                            // ===== Steigrate
					if (!user_data[INDEX_VERT_SPEED].valid)
					{
						// Die MSB Auflsung von 0,1 m/s wird auf 1 cm/s umgerechnet.
						// Hinweis:
						// OpenTX konfiguriert die Anzeige in m/s mit einer Nachkommastelle!
						user_data[INDEX_VERT_SPEED].value = msb_data[i].value * 10;
						set_user_data_valid(INDEX_VERT_SPEED, i);
					}
					break;
			}
		}
	}

	// Die zustzlichen User Daten (mit IDs gem MSB Adresse) werden berechnet.

	for (i = 0; i < 16; i++)
	{
		if (!msb_used[i] && msb_data[i].valid)
		{
			if (msb_data[i].class == MSB_RPM)
			{
				// Fr Drehzahlen gibt es zwei mgliche Auflsungen (10 bzw. 100 1/min).
				// Beide Formate werden auf 1/min mit positivem Vorzeichen umgerechnet.

				uint16_t rpm = 0;

				if (msb_data[i].value < 0)
				{
					// MSB Auflsung 10 1/min (durch negatives Vorzeichen gekennzeichnet)
					rpm = -msb_data[i].value * 10;
				}
				else
				{
					// MSB Auflsung 100 1/min (durch positives Vorzeichen gekennzeichnet)
					rpm = msb_data[i].value * 100;
				}
				if (rpm > 32767)    // Begrenzung auf maximalen positiven Wert fr User Daten
				{
					rpm = 32767;
				}
				extra_data[i].value = rpm;
			}
			else
			{
				// Alle anderen Sensorwerte werden einfach unverndert durchgereicht.
				extra_data[i].value = msb_data[i].value;
			}
			extra_data[i].valid = true;
			msb_used[i] = true;         // verwendete MSB Adresse kennzeichnen
			if (msb_data[i].updated)    // ggf. MSB Update-Flag bernehmen und dann zurcksetzen
			{
				extra_data[i].updated = true;
				msb_data[i].updated = false;
			}
		}
		else
		{
			extra_data[i].valid = false;
		}
	}

	// Die MSB Alarm-Flags werden abgefragt.

	alarm_flags = 0;
	for (i = 0; i < 16; i++)
	{
		if (msb_data[i].valid && msb_data[i].alarm)
		{
			alarm_flags |= (1 << i);
		}
	}
}

void handle_frsky_tx_request(uint8_t count)
{
	static bool rvlq_sent = false;      // zeigt an ob bereits ein RVLQ Frame gesendet wurde.
	static bool alarms_sent = false;    // zeigt an ob bereits die Alarm Flags gesendet wurden

	uint8_t i;

	// Die Flags fr RVLQ Frame und Alarm Flags werden vor dem ersten Frame zurckgesetzt.
	if (count == 0)
	{
		rvlq_sent = false;
		alarms_sent = false;
	}

	// Frame fr FrSky User Daten (falls ein Parameter aktualisiert wurde)
	for (i = 0; i < USER_COUNT; i++)
	{
		if (user_data[i].updated)
		{
			send_frsky_user_frame(user_data[i].id, user_data[i].value);
			user_data[i].updated = false;
			return;
		}
	}
	// Frame fr zustzliche User Daten (falls ein Parameter aktualisiert wurde)
	for (i = 0; i < 16; i++)
	{
		if (extra_data[i].updated)
		{
			send_frsky_user_frame(extra_data[i].id, extra_data[i].value);
			extra_data[i].updated = false;
			return;
		}
	}
	if (!(rvlq_sent))
	{
		// RVLQ Frame (wird in jeder Sequenz maximal einmal gesendet)
		send_frsky_rvlq_frame(a1_voltage, a2_voltage, rssi);
		rvlq_sent = true;
		return;
	}
	if (!(alarms_sent))
	{
		// Frame fr MSB Alarm-Flags (wird in jeder Sequenz maximal einmal gesendet)
		send_frsky_user_frame(ID_ALARMS, alarm_flags);
		alarms_sent = true;
	}
}