/***************************************************************************************************
Interface zum Senden von FrSky Telemetriedaten
****************************************************************************************************
Controller:     ATmega328P
Taktfrequenz:   16 MHz
Entwickler:     Reinhardt Werbik
Version:        1.00
Datum:          29.10.2017
****************************************************************************************************
Das FrSky Interface sendet auf Anforderung FrSky Telemetriedaten.
Es kann ein RVLQ Frame oder ein User Data Frame getriggert werden.
Der Sendepuffer wird vor dem eigentlichen Senden der Daten entsprechend beschrieben.

In einem User Data Frame wird fr die eigentlichen User Daten beim Schreiben in den Sendepuffer
Byte Stuffing nach folgendem Schema durchgefhrt.

   0x5E -> 0x5D 0x3E
   0x5D -> 0x5D 0x3D

Die Gesamtzahl der zu sendenden Bytes ndert sich durch dieses Byte Stuffing nicht.
Es beeinflusst lediglich die Anzahl gltiger Bytes im Frame.

Zum Senden der Daten ist ein SW UART mit invertierter Logik implementiert (9600 Baud).
An einem zustzlichen Ausgang wird ein UART-Signal mit normalen Pegeln zur Verfgung gestellt.

Timer 2 erzeugt mit dem Compare Match A Interrupt den Bit-Takt fr den SW UART.
Der Timer wird dazu im CTC Mode mit Prescaler 8 (Vergleichswert A = 207) betrieben.
(Taktzeit = (207 + 1) * 8 / 16 MHz = 104 s)

Beim Senden wird Byte Stuffing fr den ganzen Frame nach folgendem Schema durchgefhrt.

   0x7E -> 0x7D 0x5E
   0x7D -> 0x7D 0x5D

Dieses Byte Stuffing beeinflusst die Gesamtzahl der zu sendenden Bytes.
Die Lnge des Sendepuffers bercksichtigt allerdings keine zustzlichen Bytes.
Vielmehr wird das Byte Stuffing whrend des eigentlichen Sendevorgangs
beim Beschreiben des Puffers fr das nchste zu sendende Byte durchgefhrt.
****************************************************************************************************
nderungen:

V 1.00:
- Erste konfigurierte Version
***************************************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <stdbool.h>

#include "frsky_interface.h"

// FrSky Port-Pin

#define FRSKY_PORT  PORTD       // Port D verwenden
#define FRSKY_DDR   DDRD        // Port D DDR
#define FRSKY       PIND1       // Pin D1 fr FrSky-Ausgang (gleicher Pin wie HW UART Tx)
#define UART        PIND3       // Pin D3 fr zustzlichen UART-Ausgang mit normalen Pegeln

#define FRSKY_HIGH  FRSKY_PORT |= (1 << FRSKY)      // FrSky-Ausgang -> High
#define FRSKY_LOW   FRSKY_PORT &= ~(1 << FRSKY)     // FrSky-Ausgang -> Low
#define UART_HIGH   FRSKY_PORT |= (1 << UART)       // zustzlicher UART-Ausgang -> High
#define UART_LOW    FRSKY_PORT &= ~(1 << UART)      // zustzlicher UART-Ausgang -> Low

// FrSky Datenpaket

#define FRSKY_BYTES         11      // Anzahl der Bytes (ohne Byte Stuffing)
#define FRSKY_HEAD          0x7E    // Wert des ersten Bytes im Datenpaket
#define FRSKY_END           0x7E    // Wert des letzten Bytes im Datenpaket
#define FRSKY_PRIM_RVLQ     0xFE    // Code fr Remote Voltage and Link Quality Frame
#define FRSKY_PRIM_USER     0xFD    // Code fr User Data Frame

#define USER_HEAD           0x5E    // Wert des ersten Bytes der User Daten
#define USER_END            0x5E    // Wert des letzten Bytes der User Daten

//========== Variablen

static volatile uint8_t tx_data[FRSKY_BYTES];       // Puffer fr zu sendendes Datenpaket
static volatile uint8_t tx_byte_count = 0;          // Anzahl noch zu sendender Bytes
static volatile uint8_t tx_byte = 0;                // Puffer fr zu sendendes Byte
static volatile uint8_t tx_bit_count = 0;           // Anzahl gesendeter Bits

static volatile bool tx_byte_stuffing = false;      // Byte Stuffing fr aktuelles Byte notwendig

//========== Interne Prototypen

void trigger_frsky_tx(void);

//========== Interrupt Service Routinen

ISR(TIMER2_COMPA_vect)          // SW UART mit invertierter Logik zum Senden der FrSky Daten
{                               // (1 Start-Bit, 8 Daten-Bits, 1 Stop-Bit, kein Parity-Bit)
	switch (tx_bit_count)
	{
		case 0:     // Start-Bit
			FRSKY_HIGH;
			UART_LOW;       // zustzlicher Ausgang mit normalen Pegeln
			tx_bit_count ++;
			break;
		case 9:     // Stop-Bit
			FRSKY_LOW;
			UART_HIGH;      // zustzlicher Ausgang mit normalen Pegeln
			tx_bit_count ++;
			break;
		case 10:    // Byte komplett gesendet
			switch (tx_byte_count)
			{
				case 0:             // keine weiteren Bytes zu senden
					TCCR2B = 0;     // Bit-Takt anhalten
					break;
				case 1:             // letztes Byte
					tx_byte = tx_data[FRSKY_BYTES - tx_byte_count];
					tx_byte_stuffing = false;       // kein Byte Stuffing fr letztes Byte
					break;
				default:            // Bytes 2 bis 10
					if (tx_byte_stuffing)
					{
						tx_byte = tx_data[FRSKY_BYTES - tx_byte_count] - 0x20;
						tx_byte_stuffing = false;
					}
					else
					{
						if ((tx_data[FRSKY_BYTES - tx_byte_count] == 0x7D) || (tx_data[FRSKY_BYTES - tx_byte_count] == 0x7E))
						{
							tx_byte = 0x7D;
							tx_byte_stuffing = true;
						}
						else
						{
							tx_byte = tx_data[FRSKY_BYTES - tx_byte_count];
							tx_byte_stuffing = false;
						}
					}
			}
			if (!tx_byte_stuffing)
			{
				tx_byte_count --;
			}
			tx_bit_count = 0;
			break;
		default:                // Daten-Bit
			// Das zu sendende Bit ist das Bit 0 des Puffers fr das zu sendene Byte.
			if (tx_byte % 2 == 1)
			{
				FRSKY_LOW;      // -> Low wenn zu sendendes Bit = 1
				UART_HIGH;      // zustzlicher Ausgang mit normalen Pegeln
			}
			else
			{
				FRSKY_HIGH;     // -> High wenn zu sendendes Bit = 0
				UART_LOW;       // zustzlicher Ausgang mit normalen Pegeln
			}
			tx_byte >>= 1;                      // nchstes Bit an Position 0 schieben
			tx_bit_count ++;
	}
}

//========== Interne Funktionen

void trigger_frsky_tx(void)
{
	tx_byte = tx_data[0];               // erstes Byte holen
	tx_byte_count = FRSKY_BYTES - 1;    // Anzahl noch zu sendender Bytes setzen
	tx_byte_stuffing = false;           // kein Byte Stuffing fr erstes Byte
	tx_bit_count = 0;
	TCCR2B = (1 << CS21);               // Bit-Takt starten
}

//========== Interface Funktionen

void init_frsky_interface(void)
{
	// FrSky-Ausgang
	FRSKY_DDR |= (1 << FRSKY);      // FrSky Pin als Ausgang konfigurieren
	FRSKY_LOW;                      // auf Low setzen (Idle Zustand bei invertierter Logik)

	// Zustzlicher UART-Ausgang
	FRSKY_DDR |= (1 << UART);       // UART Pin als Ausgang konfigurieren
	UART_HIGH;                      // auf High setzen (Idle Zustand)

	// Timer 2
	TCCR2A = (1 << WGM21);          // CTC Mode auswhlen
	OCR2A = 207;                    // Vergleichswert A setzen
	TIMSK2 = (1 << OCIE2A);         // Compare Match A Interrupt freigeben

	// Sendepuffer
	uint8_t i;
	for (i = 0; i < FRSKY_BYTES; i++)
	{
		tx_data[i] = 0;
	}
}

void send_frsky_rvlq_frame(uint8_t a1, uint8_t a2, uint8_t rssi)
{
	tx_data[0] = FRSKY_HEAD;
	tx_data[1] = FRSKY_PRIM_RVLQ;
	tx_data[2] = a1;
	tx_data[3] = a2;
	tx_data[4] = rssi;
	tx_data[5] = 0x00;
	tx_data[6] = 0x00;
	tx_data[7] = 0x00;
	tx_data[8] = 0x00;
	tx_data[9] = 0x00;
	tx_data[10] = FRSKY_END;
	trigger_frsky_tx();
}

void send_frsky_user_frame(uint8_t data_id, int16_t data_value)
{
	// In dieser Funktion wird wird auch das Byte Stuffing fr die User Daten durchgefhrt.
	// Davon kann nur das Byte 6 (Low Byte des Sensorwertes) betroffen sein.

	uint8_t low_byte = data_value & 0xFF;   // Low Byte extrahieren
	uint8_t high_byte = data_value >> 8;    // High Byte extrahieren

	tx_data[0] = FRSKY_HEAD;
	tx_data[1] = FRSKY_PRIM_USER;
	tx_data[3] = 0x00;
	tx_data[4] = USER_HEAD;
	tx_data[5] = data_id;
	if ((low_byte == 0x5D) || (low_byte == 0x5E))
	{
		tx_data[2] = 6;                     // 6 gltige Bytes (Byte Stuffing notwendig)
		tx_data[6] = 0x5D;                  // zustzliches Stuffing Byte
		tx_data[7] = low_byte - 0x20;       // modifiziertes Low Byte
		tx_data[8] = high_byte;
		tx_data[9] = USER_END;
	}
	else
	{
		tx_data[2] = 5;                     // 5 gltige Bytes (kein Byte Stuffing)
		tx_data[6] = low_byte;
		tx_data[7] = high_byte;
		tx_data[8] = USER_END;
		tx_data[9] = 0x00;
	}
	tx_data[10] = FRSKY_END;
	trigger_frsky_tx();
}