/***************************************************************************************************
Interface zum Empfangen von M-Link Telemetriedaten
****************************************************************************************************
Controller:     ATmega328P
Taktfrequenz:   16 MHz
Entwickler:     Reinhardt Werbik
Version:        1.00
Datum:          30.10.2017
****************************************************************************************************
Das M-Link Interface empfngt Telemetriedaten von einem M-Link Sendemodul.

Nach der Initialisierung wird zunchst auf einen Idle Line Timeout am M-Link Eingang gewartet.
Dadurch ist sichergestellt, dass die nchste fallende Flanke der Start eines Datenpakets ist.

Zum Empfangen der Daten ist ein SW UART implementiert (115200 Baud).

Zur Detektion des Start-Bits am M-Link Eingang wird der Externe Interrupt INT0 verwendet.
Dieser wird so konfiguriert, dass der Interrupt bei fallender Flanke ausgelst wird.

Timer 0 dient nach der Initialisierung zunchst zur Idle Time berwachung am M-Link Eingang.
Der Pin Change Interrupt 2 dient zum Rcksetzen des Timers, solange kein Idle Zustand vorliegt.
Der Overflow Interrupt wird als Timeout benutzt, um den Idle Zustand zu detektieren.
Der Timer wird dazu im Normal Mode mit Prescaler 64 betrieben.
(Timeout = 256 * 64 / 16 MHz = 1,024 ms)

Danach erzeugt Timer 0 mit dem Compare Match A Interrupt den Bit-Takt fr den SW UART.
Der Timer wird dazu im CTC Mode ohne Prescaler (Vergleichswert A = 138) betrieben.
(Taktzeit = (138 + 1) / 16 MHz = 8,6875 s)

Der Timer wird nach dem Overflow Interrupt entsprechend umkonfiguriert.

Vor dem Starten des Bit-Takts zum Empfangen eines Bytes am M-Link Eingang wird der Timer vorgeladen.
Der Vorladewert wurde experimentell so ermittelt, dass der erste Interrupt zeitlich so liegt,
dass das Einlesen des ersten Daten-Bits in der Mitte des Bits erfolgt.

Die empfangenen Daten werden in einem Empfangspuffer zur Weiterverarbeitung abgelegt.
Wenn das letzte Byte eines Datenpakets empfangen wurde, wird ein entsprechendes Flag gesetzt.

Beim Empfangen wird das Byte Stuffing des Datenpakets nach folgendem Schema rckgngig gemacht.

   0x1B 0x3B -> 0x1B
   0x1B 0x22 -> 0x02
   0x1B 0x23 -> 0x03
****************************************************************************************************
nderungen:

V 1.00:
- Erste konfigurierte Version
***************************************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <stdbool.h>

#include "mlink_interface.h"

// M-Link Port-Pin

#define MLINK_PORT  PORTD       // Port D verwenden
#define MLINK_PIN   PIND        // Port D Input Pins
#define MLINK       PIND2       // Pin D2 fr M-Link Eingang (erzeugt Externen Interrupt 0)

// M-Link Datenpaket

#define MLINK_BYTES     20      // Anzahl der Bytes (ohne Byte Stuffing)
#define MLINK_HEAD      0x02    // Wert des ersten Bytes
#define MLINK_END       0x03    // Wert des letzten Bytes

// Test-Ausgang
/*
#define TEST_PORT   PORTB       // Port B verwenden
#define TEST_DDR    DDRB        // Port B DDR
#define TEST        PINB1       // Pin B1 fr Test-Ausgang

#define TEST_HIGH   TEST_PORT |= (1 << TEST)        // Test-Ausgang -> High
#define TEST_LOW    TEST_PORT &= ~(1 << TEST)       // Test-Ausgang -> Low
*/
//========== Variablen

static volatile uint8_t rx_data[MLINK_BYTES];       // Puffer fr empfangenes Datenpaket
static volatile uint8_t rx_byte_count = 0;          // Anzahl empfangener Bytes
static volatile uint8_t rx_byte = 0;                // Puffer fr einzulesendes Byte
static volatile uint8_t rx_bit_count = 0;           // Anzahl empfangener Bits

static volatile bool rx_byte_stuffing = false;      // Byte Stuffing fr aktuelles Byte notwendig
static volatile bool rx_data_complete = false;      // Flag fr komplettes Datenpaket

//========== Interrupt Service Routinen

ISR(INT0_vect)                  // Start-Bit am M-Link Eingang
{
	EIMSK &= ~(1 << INT0);      // Externen Interrupt 0 sperren
	rx_bit_count = 1;           // Start-Bit ist mitzuzhlen
	TCNT0 = 250;                // Timer fr Bit-Takt vorladen
	TCCR0B = (1 << CS00);       // Bit-Takt starten
}

ISR(PCINT2_vect)                // Pegelnderung am M-Link Eingang
{
	TCNT0 = 0;                  // Idle Time auf 0 zurcksetzen
}

ISR(TIMER0_COMPA_vect)          // SW UART zum Empfangen der M-Link Daten
{                               // (1 Start-Bit, 8 Daten-Bits, 1 Stop-Bit, kein Parity-Bit)
	if (rx_bit_count < 9)
	{
		// Mindestens ein Bit ist noch zu empfangen.
		rx_byte >>= 1;          // hchstes Bit im Puffer freimachen
		//TEST_HIGH;
		if (MLINK_PIN & (1 << MLINK))
		{
			rx_byte |= 0x80;    // und gem zu empfangendem Bit setzen
		}
		//TEST_LOW;
		rx_bit_count ++;
	}
	else
	{
		// Alle Bits wurden empfangen.
		if (MLINK_PIN & (1 << MLINK))       // Prfen ob Stop-Bit anliegt
		{
			// Das empfangene Byte wird abgelegt, wobei das Byte Stuffing rckgngig gemacht wird.
			rx_byte_count ++;
			if (rx_byte_stuffing)
			{
				rx_data[rx_byte_count -1] = rx_byte - 0x20;
				rx_byte_stuffing = false;
			}
			else
			{
				if (rx_byte == 0x1B)
				{
					rx_byte_stuffing = true;    // Daten-Byte 0x1B ignorieren
					rx_byte_count --;           // Byte-Zhler anpassen
				}
				else
				{
					rx_data[rx_byte_count -1] = rx_byte;
					rx_byte_stuffing = false;
				}
			}
			// Es wird geprft, ob das empfangene Byte das letzte eines Datenpakets ist.
			if (rx_byte == MLINK_END)
			{
				rx_data_complete = true;    // komplettes Datenpaket signalisieren
				rx_byte_count = 0;
			}
		}
		TCCR0B = 0;                 // Bit-Takt anhalten
		EIFR |= (1 << INTF0);       // Interrupt Flag fr Externen Interrupt zurcksetzen
		EIMSK |= (1 << INT0);       // Externen Interrupt 0 freigeben
	}
}

ISR(TIMER0_OVF_vect)            // Idle Line Timeout am M-Link Eingang nach Initialisierung
{
	// Idle Time berwachung deaktivieren (Timer 0)
	TCCR0B = 0;                 // Idle Time anhalten
	TIMSK0 &= ~(1 << TOIE0);    // Overflow Interrupt sperren

	// Timer 0 fr Bit-Takt umkonfigurieren
	TCCR0A = (1 << WGM01);      // CTC Mode auswhlen
	OCR0A = 138;                // Vergleichswert A setzen
	TIMSK0 |= (1 << OCIE0A);    // Compare Match A Interrupt freigeben

	// Detektion des Start-Bits ermglichen
	PCICR &= ~(1 << PCIE2);     // Pin Change Interrupt 2 sperren
	EIFR |= (1 << INTF0);       // Interrupt Flag fr Externen Interrupt zurcksetzen
	EIMSK |= (1 << INT0);       // Externen Interrupt 0 freigeben
}

//========== Interface Funktionen

void init_mlink_interface(void)
{
	// M-Link Eingang
	MLINK_PORT &= ~(1 << MLINK);    // Pull-Up fr M-Link Eingang deaktivieren
	EICRA = (1 << ISC01);           // Externer Interrupt 0 bei fallender Flanke
	PCICR = (1 << PCIE2);           // Pin Change Interrupt 2 freigeben
	PCMSK2 = (1 << PCINT18);        // Pin Change Interrupt nur durch Pin fr M-Link Eingang

	// Test-Ausgang
	//TEST_DDR |= (1 << TEST);        // Test Pin als Ausgang konfigurieren
	//TEST_LOW;

	// Timer 0
	TCCR0B = (1 << CS01) | (1 << CS00);     // Idle Time starten
	TIMSK0 = (1 << TOIE0);                  // Overflow Interrupt freigeben

	// Empfangspuffer
	uint8_t i;
	for (i = 0; i < MLINK_BYTES; i++)
	{
		rx_data[i] = 0;
	}
}

bool new_mlink_data(void)
{
	if (rx_data_complete)
	{
		rx_data_complete = false;
		return true;
	}
	else
	{
		return false;
	}
}

bool mlink_data_valid(void)
{
	// Das achte Byte des M-Link Datenpakets ist das Status-Byte.
	// Gltige Telemetriedaten werden fr folgende Werte des M-Link Status-Bytes signalisiert:
	// - normaler Betrieb ohne Fast Response    -> 0x06
	// - normaler Betrieb mit Fast Response     -> 0x04
	// - Reichweitentest ohne Fast Response     -> 0x46
	// - Reichweitentest mit Fast Response      -> 0x44
	// (Diese Werte setzen jeweils einen gebundenen telemetriefhigen M-Link Empfnger voraus.)

	if ((rx_data[7] == 0x06) || (rx_data[7] == 0x04) || (rx_data[7] == 0x46) || (rx_data[7] == 0x44))
	{
		return true;
	}
	else
	{
		return false;
	}
}

bool fast_response_enabled(void)
{
	// Im normalen Betrieb ist das zweitniedrigste Bit im Low Nibble des achten Bytes gesetzt.
	// Wenn Fast Response aktiviert ist, ist dieses Bit nicht gesetzt.

	if (rx_data[7] & 0x02)
	{
		return false;
	}
	else
	{
		return true;
	}
}

uint8_t mlink_byte(uint8_t index)
{
	return rx_data[index];
}